% [XX,ff,FF,E] = drur(Y,X,fmap,Fmap[,tol,maxit])
% Dimensionality reduction by unsupervised regression (DRUR)
%
% Typical use with RBFs:
%   [XX,ff,FF,E] = drur(Y,X,{'rbf',70,[],1e-5},{'rbf',70,[],1e-5},[],30);
% If you want to give specific initial values to the RBF centres or fix their
% width, note carefully how 'rbftrain' is called below (with drurnorm=1 it
% makes no sense to fix the width for f).
%
% In:
%   Y: NxD matrix containing N D-dim data points rowwise (high-dim data).
%   X: NxL matrix containing N L-dim data points rowwise (initial low-dim
%      points, obtained from a spectral method).
%   fmap, Fmap: cell arrays describing the type of mapping (f:X->Y, F:Y->X):
%    {'rbf',C,s,l}: RBF network with parameters C,s,l as in 'rbftrain'.
%    {'mlp',H}: single-layer MLP with H sigmoidal hidden units.
%    {'linf',l}: linear mapping with regularisation parameter l.
%    {'nonp',sx,sy,lf,lF}: nonparametric DRUR.
%   tol: minimum relative decrease in error to keep iterating. Default: 1e-4.
%   maxit: maximum number of iterations. Default: 1000.
% Out: each of the following is a cell array where element i+1 corresponds to
%   the ith-iteration:
%   XX: NxL matrix containing N L-dim data points rowwise.
%   ff: (struct) parameters for the forward mapping according to f.type:
%    'rbf': centres C (MxL), width s, weights W (Dx(M+1), incl. bias), reg. l.
%    'mlp': weight matrices (incl. bias) layers 1/2: W1 (Hx(L+1)),W2 (Dx(H+1)).
%    'linf': weight matrix W (DxL), bias vector w (Dx1), reg. l.
%   FF: as for f but for the backward mapping.
%   E: error (scalar).
%
% Any non-mandatory argument can be given the value [] to force it to take
% its default value.

% Copyright (c) 2010 by Miguel A. Carreira-Perpinan

function [XX,ff,FF,E] = drur(Y,X,fmap,Fmap,tol,maxit)

% ---------- Argument defaults ----------
if ~exist('tol','var') | isempty(tol) tol = 1e-4; end;
if ~exist('maxit','var') | isempty(maxit) maxit = 1000; end;
for i=length(fmap)+1:4 fmap{i} = []; end;
for i=length(Fmap)+1:4 Fmap{i} = []; end;
% ---------- End of "argument defaults" ----------

fcn = str2func(fmap{1});	% Use string as function handle

rbfit = 10;	% Update RBF centres/width only every rbfit iterations...
rbfmaxit = 53;	% ...and if i < rbfmaxit

% drurnorm = 1 normalises X and Y to zero mean and unit Frobenius norm. This
% equalises their errors and prevents shrinkage of X over DRUR iterations.
% The main DRUR loop works with the normalised X and Y, renormalising X after
% each Xopt. We unnormalise back f and F (wrt Y, not X) before saving them to
% ff{} and FF{}, so the resulting mappings apply directly to the original Y.
% 'linf' is much faster with drurnorm = 1.
drurnorm = 1;	% drurnorm is currently only valid for rbf and linf, not mlp

if drurnorm
  mY = mean(Y,1); Y = bsxfun(@minus,Y,mY); sY = norm(Y,'fro'); Y = Y/sY;
  mX = mean(X,1); X = bsxfun(@minus,X,mX); sX = norm(X,'fro'); X = X/sX;
end

% Fit mappings f, F
% Variables Ef2, EF2 are needed to account for the RBF regularisation error
fprintf('DRUR iteration  %3d:\n',0);
fprintf('  training f: ');
switch fmap{1}
 case 'rbf', [f,fX,Ef] = rbftrain(X,Y,fmap{2:4}); Ef2 = Ef(2); Ef = sum(Ef);
 case 'linf', [f,fX,Ef] = linftrain(X,Y,fmap{2}); Ef2 = Ef(2); Ef = sum(Ef);
 case 'mlp', [f,fX,Ef] = mlptrain(X,Y,fmap{2},[],1e-9,1e5); Ef=Ef(end); Ef2=0;
end
fprintf('f-error=%g\n  training F: ',Ef);
switch Fmap{1}
 case 'rbf', [F,FY,EF] = rbftrain(Y,X,Fmap{2:4}); EF2 = EF(2); EF = sum(EF);
 case 'linf', [F,FY,EF] = linftrain(Y,X,Fmap{2}); EF2 = EF(2); EF = sum(EF);
 case 'mlp', [F,FY,EF] = mlptrain(Y,X,Fmap{2},[],1e-9,1e5); EF=EF(end); EF2=0;
end
fprintf('F-error=%g\n',EF);
E = Ef+EF; fprintf('              DRUR error=%g\n',E);

i = 0; cont = (maxit>=1); ff{1} = f; FF{1} = F; XX{1} = X;

while cont
  
  fprintf('DRUR iteration  %3d:\n',i+1);

  % PROJECTION: optimisation over X
  fprintf('  training X: ');
  X = Xopt(X,Y,f,FY);
  
  if drurnorm
    % Normalise X, f, FY (don't need to normalise F since it's adapted below)
    mX = mean(X,1); X = bsxfun(@minus,X,mX); sX = norm(X,'fro'); X = X/sX;
    switch fmap{1}
     case 'rbf', f.C = bsxfun(@minus,f.C,mX); f.C = f.C/sX; f.s = f.s/sX;
     case 'linf', f.w = f.W*mX'+f.w; f.W = f.W*sX;
     case 'mlp', 
    end
    FY = bsxfun(@minus,FY,mX); FY = FY/sX; EF2 = EF2/(sX*sX);
  end
  
  fX = fcn(X,f);
  EfX = sum(sum((Y-fX).^2))+Ef2; EFX = sum(sum((X-FY).^2))+EF2; EX = EfX+EFX;
  fprintf('DRUR error=%g (f-error=%g, F-error=%g)\n',EX,EfX,EFX);

  % ADAPTATION: optimisation over f, F
  fprintf('  training f: ');
  switch fmap{1}
   case 'rbf',
    if isempty(fmap{3}) & (i<rbfmaxit) & (mod(i,rbfit)==0) f.s=f.s*[1/2 2]; end
    [f,fX,Ef] = rbftrain(X,Y,f.C,f.s,fmap{4}); Ef2 = Ef(2); Ef = sum(Ef);
   case 'linf', [f,fX,Ef] = linftrain(X,Y,fmap{2}); Ef2 = Ef(2); Ef = sum(Ef);
   case 'mlp', [f,fX,Ef] = mlptrain(X,Y,f); Ef = Ef(end);
  end
  fprintf('DRUR error=%g (f-error=%g)\n  training F: ',EX-EfX+Ef,Ef);
  switch Fmap{1}
   case 'rbf',
    [F,FY,EF] = rbftrain(Y,X,F.C,F.s,Fmap{4}); EF2 = EF(2); EF = sum(EF);
   case 'linf', [F,FY,EF] = linftrain(Y,X,Fmap{2}); EF2 = EF(2); EF = sum(EF);
   case 'mlp', [F,FY,EF] = mlptrain(Y,X,F,[],[],1e4); EF = EF(end);
  end
  fprintf('DRUR error=%g (F-error=%g)\n',Ef+EF,EF);

  E = [E Ef+EF];		% Error  
  i = i + 1;
  ff{i+1} = f; FF{i+1} = F; XX{i+1} = X;
  
  % Allow error to increase occasionally because of heuristic rbftrain for f
  % or if drurnorm=1:
  cont = (abs(E(end-1)-E(end))>tol*E(end-1)) & (i<maxit);
  % cont = ((E(end-1)-E(end))>tol*E(end-1)) & (i<maxit);
  % cont = (((E(end-1)-E(end))>tol*E(end-1)) | (E(end-1)<E(end))) & (i<maxit);
  
end

if drurnorm
  % Unnormalise back f and F (wrt Y, not X)
  for i=1:length(ff)
    Fu = FF{i}; fu = ff{i};
    switch fmap{1}
     case 'rbf',
      fu.W(:,1:end-1) = fu.W(:,1:end-1)*sY; fu.W(:,end) = fu.W(:,end)*sY + mY';
     case 'linf', fu.W = fu.W*sY; fu.w = fu.w*sY+mY';
     case 'mlp', 
    end
    switch Fmap{1}
     case 'rbf', Fu.C = Fu.C*sY; Fu.C = bsxfun(@plus,Fu.C,mY); Fu.s = Fu.s*sY;
     case 'linf', Fu.W = Fu.W/sY; Fu.w = -Fu.W*mY'+Fu.w;
     case 'mlp', 
    end
    ff{i} = fu; FF{i} = Fu;
  end
end

fprintf('.\n');


% [X,EE,AA] = Xopt(X,Y,f,FY[,method,tol,maxit])
% Optimise reconstruction error wrt low-dim coordinate x:
%   min_x { |y-f(x)| + |x-F(y)| }
% independently for each vector x = X(n,:), y = Y(n,:).
%
% In:
%   X: NxL matrix, initial values.
%   Y: NxD matrix.
%   f: (struct) mapping f:x->y.
%   FY: NxL matrix, F(Y).
%   method: 'grad', 'fp' or 'GN'. Default: 'GN'.
%   tol: minimum relative decrease in error to keep iterating. Default: 1e-4.
%   maxit: maximum number of iterations. Default: 1000.
% Out:
%   X: NxL matrix, approx. optimal values.
%   EE: cell array with N elements, each a sequence of the reconstruction
%      errors during the optimisation.
%   AA: cell array with N elements, each a sequence of the step lengths
%      during the optimisation.

% Copyright (c) 2008 by Miguel A. Carreira-Perpinan

function [X,EE,AA] = Xopt(X,Y,f,FY,method,tol,maxit)

% ---------- Argument defaults ----------
if ~exist('method','var') | isempty(method) method = 'GN'; end;
if ~exist('tol','var') | isempty(tol) tol = 1e-4; end;
if ~exist('maxit','var') | isempty(maxit) maxit = 1000; end;
% ---------- End of "argument defaults" ----------

[N,L] = size(X);
fcn = str2func(f.type);		% Use string as function handle
eta = 1e-2;			% Step size for 'grad'

% For each x separately
for n=1:N
  
  x = X(n,:); y = Y(n,:); Fy = FY(n,:); [fx,J] = fcn(x,f);
  
  cont = (maxit>=1); i = 0; E = (y-fx)*(y-fx)'+(x-Fy)*(x-Fy)'; A = [];
  
  while cont
    
    g = 2*(x - Fy - (y-fx)*J);
    
    switch method
     case 'grad'
      x = x - eta*g;
      alpha = eta;
     case 'fp'
      x = Fy + (y-fx)*J;
      alpha = 0.5;
     case 'GN'
      p = -( (eye(L) + J'*J) \ g' )/2;
      e = (y-fx)*(y-fx)'+(x-Fy)*(x-Fy)';
      % Backtracking line search
      alpha = linesearch(@xerror,{y,f,Fy},x',p,e,g);
      x = x + alpha*p';
    end
    [fx,J] = fcn(x,f);
    
    E = [E (y-fx)*(y-fx)'+(x-Fy)*(x-Fy)'];
    A = [A alpha];
    i = i + 1;
    cont = (((1-tol)*E(end-1)>E(end))) & (i<maxit);
    
  end
  
  X(n,:) = x;
  if nargout>1 EE{n} = E; end
  if nargout>2 AA{n} = A; end

end


% e = xerror(x,y,f,Fy) Reconstruction error over x: e = |y-f(x)| + |x-Fy|

function e = xerror(x,y,f,Fy)

fcn = str2func(f.type);		% Use string as function handle
e = sum([y-fcn(x,f) x-Fy].^2);


% alpha = linesearch(f,paramf,x,p,ff,g[,alpha0,rho,c]) Backtracking line search
%
% Reference: procedure 3.1, p. 41ff in:
%   Nocedal and Wright: "Numerical Optimization", Springer, 1999.
%
% In:
%   f, paramf: objective function handle and its parameters.
%   x: Lx1 vector, the current iterate.
%   p: Lx1 vector, the search direction.
%   ff: value at x of the function.
%   g: 1xL vector, gradient at x of the function.
%   alpha0: initial step size. Default: 1.
%   rho: rate of decrease of the step size. Default: 0.8.
%   c: Wolfe condition. Default: 1e-4.
% Out:
%   alpha: step size.

% Copyright (c) 2006 by Miguel A. Carreira-Perpinan

function alpha = linesearch(f,paramf,x,p,ff,g,alpha0,rho,c)

% ---------- Argument defaults ----------
if ~exist('alpha0','var') | isempty(alpha0) alpha0 = 1; end;
if ~exist('rho','var') | isempty(rho) rho = 0.8; end;
if ~exist('c','var') | isempty(c) c = 1e-4; end;
% ---------- End of "argument defaults" ----------
maxit = 100;

alpha = alpha0;
tmp = c*g*p;
if tmp > -1e5*eps(ff)	% Precision loss, can't reduce f anymore
  alpha = 0;
else
  j = 0;		% Limit iterations just in case
  while (f((x+alpha*p)',paramf{:}) > ff+alpha*tmp) & (j<maxit)
    alpha = rho*alpha;
    j = j + 1;
  end
  if f((x+alpha*p)',paramf{:}) >= ff alpha = 0; end
  if j==maxit fprintf('maxit'); end
end

