% [f,fX,E] = mlptrain(X,Y,f[,eta,tol,maxit]) Train MLP to map f(X) = Y
%
% The MLP we use has a single hidden layer of H sigmoidal units.
%
% In:
%   X: NxL matrix, N L-dim data points rowwise.
%   Y: NxD matrix, N D-dim data points rowwise.
%   f: if a scalar, the #hiddens to use;
%      else, (struct) the MLP, initial values for the weights and biases.
%   eta: backprop learning rate (gradient step size). Default: 2e-5.
%   tol: minimum relative decrease in error to keep iterating. Default: 1e-6.
%   maxit: maximum number of iterations. Default: 1e5.
% Out:
%   f: (struct) the MLP, with fields:
%      type='mlp', weight matrices (incl. bias) layers 1/2: W1 (Hx(L+1)),
%      W2 (Dx(H+1)).
%   fX: NxD matrix, f(X).
%   E: sequence of reconstruction errors during the optimisation.

% Copyright (c) 2009 by Miguel A. Carreira-Perpinan

function [f,fX,E] = mlptrain(X,Y,f,eta,tol,maxit)

% ---------- Argument defaults ----------
if ~exist('eta','var') | isempty(eta) eta = 2e-5; end;
if ~exist('tol','var') | isempty(tol) tol = 1e-6; end;
if ~exist('maxit','var') | isempty(maxit) maxit = 1e5; end;
% ---------- End of "argument defaults" ----------

[N,L] = size(X); D = size(Y,2);

if ~isstruct(f)
  H = f; clear f;
  % Initialise weights and biases to small random values
  f.W1 = randn(H,L+1)/10; f.W2 = randn(D,H+1)/10;
end;
f.type = 'mlp';

cont = (maxit>=1); i = 0; fX = mlp(X,f); e = (fX-Y)'; E = e(:)'*e(:);
while cont
  
  % Gradients
  [s,ds] = sigmoid(f.W1*[X';ones(1,N)]);
  gW2 = [e*s' sum(e,2)];
  ee = ds.*(f.W2(:,1:end-1)'*e);
  gW1 = [ee*X sum(ee,2)];
  
  % Gradient step
  f.W2 = f.W2 - 2*eta*gW2; f.W1 = f.W1 - 2*eta*gW1;
  
  fX = mlp(X,f); e = (fX-Y)'; E = [E e(:)'*e(:)];
  i = i + 1;
  cont = ((E(end-1)-E(end))>tol*E(end-1)) & (i<maxit);
  fprintf('\r  mlptrain iteration: %3d; error: %f',i,E(end));

end

if E(end)>E(end-1) fprintf(' ***overshoot'); end;
fprintf('.\n');

if nargout>1 fX = mlp(X,f); end

